/* This file is autogenerated by build-utils/gentokens.lua */
#ifndef LUAKIT_COMMON_TOKENIZE_H
#define LUAKIT_COMMON_TOKENIZE_H

#include <glib.h>

typedef enum luakit_token_t {
    L_TK_UNKNOWN=0, /* (luakit_token_t) 0 == L_TK_UNKNOWN */
    L_TK_ACCEPT_POLICY,
    L_TK_ADD_EVENT_LISTENER,
    L_TK_ALIGN,
    L_TK_ALLOW_CERTIFICATE,
    L_TK_ALLOW_FILE_ACCESS_FROM_FILE_URLS,
    L_TK_ALLOW_MODAL_DIALOGS,
    L_TK_ALLOW_OVERWRITE,
    L_TK_ALLOW_UNIVERSAL_ACCESS_FROM_FILE_URLS,
    L_TK_APPEND,
    L_TK_ATTR,
    L_TK_AUTO_LOAD_IMAGES,
    L_TK_BASELINE,
    L_TK_BG,
    L_TK_BODY,
    L_TK_BOTTOM,
    L_TK_CACHE_DIR,
    L_TK_CAN_FOCUS,
    L_TK_CAN_GO_BACK,
    L_TK_CAN_GO_FORWARD,
    L_TK_CENTER,
    L_TK_CERTIFICATE,
    L_TK_CHECKED,
    L_TK_CHILD,
    L_TK_CHILDREN,
    L_TK_CHILD_COUNT,
    L_TK_CLEAR,
    L_TK_CLEAR_SEARCH,
    L_TK_CLICK,
    L_TK_CLIENT_RECTS,
    L_TK_CLIPBOARD,
    L_TK_CLOSE_INSPECTOR,
    L_TK_CONFIG_DIR,
    L_TK_CONFPATH,
    L_TK_COOKIES_STORAGE,
    L_TK_COUNT,
    L_TK_CRASH,
    L_TK_CREATE_ELEMENT,
    L_TK_CSS,
    L_TK_CURRENT,
    L_TK_CURRENT_SIZE,
    L_TK_CURSIVE_FONT_FAMILY,
    L_TK_DATA_DIR,
    L_TK_DECORATED,
    L_TK_DEFAULT_CHARSET,
    L_TK_DEFAULT_FONT_FAMILY,
    L_TK_DEFAULT_FONT_SIZE,
    L_TK_DEFAULT_MONOSPACE_FONT_SIZE,
    L_TK_DESKTOP_DIR,
    L_TK_DESTINATION,
    L_TK_DESTROY,
    L_TK_DEV_PATHS,
    L_TK_DOCUMENT,
    L_TK_DOCUMENTS_DIR,
    L_TK_DOWNLOAD_DIR,
    L_TK_DRAWING_AREA,
    L_TK_DRAW_COMPOSITING_INDICATORS,
    L_TK_EDITABLE,
    L_TK_ELAPSED_TIME,
    L_TK_ELEMENT_FROM_POINT,
    L_TK_ENABLE_ACCELERATED_2D_CANVAS,
    L_TK_ENABLE_CARET_BROWSING,
    L_TK_ENABLE_DEVELOPER_EXTRAS,
    L_TK_ENABLE_DNS_PREFETCHING,
    L_TK_ENABLE_FRAME_FLATTENING,
    L_TK_ENABLE_FULLSCREEN,
    L_TK_ENABLE_HTML5_DATABASE,
    L_TK_ENABLE_HTML5_LOCAL_STORAGE,
    L_TK_ENABLE_HYPERLINK_AUDITING,
    L_TK_ENABLE_JAVA,
    L_TK_ENABLE_JAVASCRIPT,
    L_TK_ENABLE_MEDIASOURCE,
    L_TK_ENABLE_MEDIA_STREAM,
    L_TK_ENABLE_PAGE_CACHE,
    L_TK_ENABLE_PLUGINS,
    L_TK_ENABLE_RESIZABLE_TEXT_AREAS,
    L_TK_ENABLE_SCRIPTS,
    L_TK_ENABLE_SITE_SPECIFIC_QUIRKS,
    L_TK_ENABLE_SMOOTH_SCROLLING,
    L_TK_ENABLE_SPATIAL_NAVIGATION,
    L_TK_ENABLE_SPELL_CHECKING,
    L_TK_ENABLE_TABS_TO_LINKS,
    L_TK_ENABLE_WEBAUDIO,
    L_TK_ENABLE_WEBGL,
    L_TK_ENABLE_WRITE_CONSOLE_MESSAGES_TO_STDOUT,
    L_TK_ENABLE_XSS_AUDITOR,
    L_TK_END,
    L_TK_ENTRY,
    L_TK_ERROR,
    L_TK_EVAL_JS,
    L_TK_EVENTBOX,
    L_TK_EXECPATH,
    L_TK_FANTASY_FONT_FAMILY,
    L_TK_FETCH,
    L_TK_FG,
    L_TK_FILENAME,
    L_TK_FILL,
    L_TK_FINISHED,
    L_TK_FIRST_CHILD,
    L_TK_FOCUS,
    L_TK_FOCUSED,
    L_TK_FONT,
    L_TK_FULLSCREEN,
    L_TK_GET_SOURCE,
    L_TK_GET_TITLE,
    L_TK_GO_BACK,
    L_TK_GO_FORWARD,
    L_TK_HARDWARE_ACCELERATION_POLICY,
    L_TK_HBOX,
    L_TK_HEIGHT,
    L_TK_HIDE,
    L_TK_HISTORY,
    L_TK_HOMOGENEOUS,
    L_TK_HOVERED_URI,
    L_TK_HPANED,
    L_TK_HREF,
    L_TK_ICON,
    L_TK_ID,
    L_TK_IMAGE,
    L_TK_INDEXOF,
    L_TK_INNER_HEIGHT,
    L_TK_INNER_HTML,
    L_TK_INNER_WIDTH,
    L_TK_INSERT,
    L_TK_INSPECTOR,
    L_TK_INSTALL_PATH,
    L_TK_INSTALL_PATHS,
    L_TK_INTERVAL,
    L_TK_INVALIDATE,
    L_TK_IS_ALIVE,
    L_TK_IS_LOADING,
    L_TK_IS_PLAYING_AUDIO,
    L_TK_JAVASCRIPT_CAN_ACCESS_CLIPBOARD,
    L_TK_JAVASCRIPT_CAN_OPEN_WINDOWS_AUTOMATICALLY,
    L_TK_LABEL,
    L_TK_LAST_CHILD,
    L_TK_LEFT,
    L_TK_LOADING,
    L_TK_LOAD_STRING,
    L_TK_MARGIN,
    L_TK_MARGIN_BOTTOM,
    L_TK_MARGIN_LEFT,
    L_TK_MARGIN_RIGHT,
    L_TK_MARGIN_TOP,
    L_TK_MAXIMIZED,
    L_TK_MEDIA_PLAYBACK_ALLOWS_INLINE,
    L_TK_MEDIA_PLAYBACK_REQUIRES_GESTURE,
    L_TK_MIME_TYPE,
    L_TK_MINIMUM_FONT_SIZE,
    L_TK_MIN_SIZE,
    L_TK_MONOSPACE_FONT_FAMILY,
    L_TK_MUSIC_DIR,
    L_TK_NAME,
    L_TK_NEXT_SIBLING,
    L_TK_NOTEBOOK,
    L_TK_NOUNIQUE,
    L_TK_OPTIONS,
    L_TK_OVERLAY,
    L_TK_OWNER_DOCUMENT,
    L_TK_PACK,
    L_TK_PACK1,
    L_TK_PACK2,
    L_TK_PARENT,
    L_TK_PATTERN,
    L_TK_PICTOGRAPH_FONT_FAMILY,
    L_TK_PICTURES_DIR,
    L_TK_PLUGGED,
    L_TK_POSITION,
    L_TK_PREV_SIBLING,
    L_TK_PRIMARY,
    L_TK_PRINT_BACKGROUNDS,
    L_TK_PRIVATE,
    L_TK_PROCESS_LIMIT,
    L_TK_PROGRESS,
    L_TK_PROXY_URI,
    L_TK_PUBLIC_SHARE_DIR,
    L_TK_QUERY,
    L_TK_RECT,
    L_TK_RELOAD,
    L_TK_RELOAD_BYPASS_CACHE,
    L_TK_REMOVE,
    L_TK_REMOVE_EVENT_LISTENER,
    L_TK_REORDER,
    L_TK_REPLACE,
    L_TK_RESOURCE_PATH,
    L_TK_RIGHT,
    L_TK_ROOT_WIN_XID,
    L_TK_SANS_SERIF_FONT_FAMILY,
    L_TK_SAVE,
    L_TK_SCALE,
    L_TK_SCREEN,
    L_TK_SCROLL,
    L_TK_SCROLLBARS,
    L_TK_SCROLLED,
    L_TK_SCROLL_X,
    L_TK_SCROLL_Y,
    L_TK_SEARCH,
    L_TK_SEARCH_NEXT,
    L_TK_SEARCH_PREVIOUS,
    L_TK_SECONDARY,
    L_TK_SELECTABLE,
    L_TK_SELECTION,
    L_TK_SELECT_REGION,
    L_TK_SEND_KEY,
    L_TK_SERIF_FONT_FAMILY,
    L_TK_SESSION_STATE,
    L_TK_SET_DARK_MODE,
    L_TK_SET_DEFAULT_SIZE,
    L_TK_SET_FAVICON_FOR_URI,
    L_TK_SET_PDFJS,
    L_TK_SET_TITLE,
    L_TK_SHOW,
    L_TK_SHOW_BORDER,
    L_TK_SHOW_FRAME,
    L_TK_SHOW_INSPECTOR,
    L_TK_SHOW_TABS,
    L_TK_SOCKET,
    L_TK_SOURCE,
    L_TK_SPACING,
    L_TK_SPELL_CHECKING_LANGUAGES,
    L_TK_SPINNER,
    L_TK_SRC,
    L_TK_SSL_TRUSTED,
    L_TK_STACK,
    L_TK_START,
    L_TK_STARTED,
    L_TK_STATUS,
    L_TK_STOP,
    L_TK_STYLE,
    L_TK_STYLESHEETS,
    L_TK_SUBMIT,
    L_TK_SUGGESTED_FILENAME,
    L_TK_SWITCH,
    L_TK_SYSTEM_CONFIG_DIRS,
    L_TK_SYSTEM_DATA_DIRS,
    L_TK_TAG_NAME,
    L_TK_TEMPLATES_DIR,
    L_TK_TEXT,
    L_TK_TEXTWIDTH,
    L_TK_TEXT_CONTENT,
    L_TK_TITLE,
    L_TK_TOOLTIP,
    L_TK_TOP,
    L_TK_TOTAL_SIZE,
    L_TK_TYPE,
    L_TK_URGENCY_HINT,
    L_TK_URI,
    L_TK_USER_AGENT,
    L_TK_VALUE,
    L_TK_VBOX,
    L_TK_VERBOSE,
    L_TK_VERSION,
    L_TK_VIDEOS_DIR,
    L_TK_VISIBLE,
    L_TK_VISIBLE_CHILD,
    L_TK_VPANED,
    L_TK_WEBKIT2,
    L_TK_WEBKIT_USER_AGENT_VERSION,
    L_TK_WEBKIT_VERSION,
    L_TK_WEBSITE_DATA,
    L_TK_WEBVIEW,
    L_TK_WEB_PROCESS_ID,
    L_TK_WIDTH,
    L_TK_WINDOW,
    L_TK_WINDOWS,
    L_TK_WIN_XID,
    L_TK_WRAP_JS,
    L_TK_X,
    L_TK_XMAX,
    L_TK_XPAGE_SIZE,
    L_TK_Y,
    L_TK_YMAX,
    L_TK_YPAGE_SIZE,
    L_TK_ZOOM_LEVEL,
    L_TK_ZOOM_TEXT_ONLY,
} luakit_token_t;

__attribute__((pure)) enum luakit_token_t l_tokenize(const gchar *);
__attribute__((pure)) const gchar * token_tostring(luakit_token_t);

#endif
